# Response class
from flask import Response


class SAFRSResponse(Response):
    """
    Response class
    """

    safrs_headers = {"Content-Type": "application/vnd.api+json"}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



# JSON:API response formatting functions:
# - filtering (https://jsonapi.org/format/#fetching-filtering)
# - sorting (https://jsonapi.org/format/#fetching-sorting)
# - pagination (https://jsonapi.org/format/#fetching-pagination)
#
# Response formatting follows filter -> sort -> paginate
#
import sqlalchemy
import sqlalchemy.orm.dynamic
import sqlalchemy.orm.collections
import safrs
from flask import request
from .jsonapi_attr import is_jsonapi_attr
from .errors import ValidationError, GenericError
from .config import get_config, get_request_param


def jsonapi_filter_list(relation):
    """
    :param relation: InstrumentedList
    :return: list of instances filtered using the jsonapi filters in the url query args

    Called when filtering a relationship collection
    """
    result = set()
    for instance in relation:
        if not hasattr(instance, "id_type"):
            # item is not a SAFRSBase instance
            result.add(instance)
            continue
        pks = {col.name: getattr(instance, col.name) for col in instance.id_type.columns}
        filter_query = instance.__class__.jsonapi_filter()
        result.update(filter_query.filter_by(**pks).all())  # this should only contain zero or one items
    return list(result)


def jsonapi_filter_query(object_query, safrs_object):
    """
    :param object_query: query to be filtered (lazy='dynamic' relationships AppenderBaseQuery)
    :param safrs_object: sqla object to be queried
    :return: sqla query object

    Called when filtering a relationship query
    """
    filter_query = safrs_object.jsonapi_filter()
    result = object_query.intersect(filter_query)
    return result


def jsonapi_sort(object_query, safrs_object):
    """
    http://jsonapi.org/format/#fetching-sorting
    sort by csv sort= values
    :param object_query: sqla query object
    :param safrs_object: SAFRSObject
    :return: sqla query object
    """
    sort_attrs = request.args.get("sort", "") or "id"

    for sort_attr in sort_attrs.split(","):
        reverse = sort_attr.startswith("-")
        if reverse:
            # if the sort column starts with - , then we want to do a reverse sort
            # The sort order for each sort field MUST be ascending unless it is prefixed
            # with a minus, in which case it MUST be descending.
            sort_attr = sort_attr[1:]
            attr = getattr(safrs_object, sort_attr, None)
            if attr is not None:
                attr = attr.desc()
        else:
            attr = getattr(safrs_object, sort_attr, None)
        if sort_attr == "id":
            if attr is None:
                if safrs_object.id_type.primary_keys:
                    attr = getattr(safrs_object, safrs_object.id_type.primary_keys[0])  # todo: composite keys edge case
                else:
                    continue
        elif attr is None or sort_attr not in safrs_object._s_jsonapi_attrs:
            safrs.log.debug(f"{safrs_object} has no attribute {sort_attr} in {safrs_object._s_jsonapi_attrs}")
            continue
        if isinstance(object_query, (list, sqlalchemy.orm.collections.InstrumentedList)):
            object_query = sorted(
                list(object_query), key=lambda obj: (getattr(obj, sort_attr) is None, getattr(obj, sort_attr)), reverse=reverse
            )
        elif is_jsonapi_attr(attr):
            # to do: implement sorting for jsonapi_attr
            safrs.log.debug(f"sorting not implemented for {attr}")
        elif hasattr(object_query, "order_by"):
            try:
                # This may fail on non-sqla objects, eg. properties
                object_query = object_query.order_by(attr)
            except sqlalchemy.exc.ArgumentError as exc:
                safrs.log.warning(f"Sort failed for {safrs_object}.{sort_attr}: {exc}")
            except Exception as exc:
                safrs.log.warning(f"Sort failed for {safrs_object}.{sort_attr}: {exc}")

    return object_query


def paginate(object_query, SAFRSObject=None):
    """
    this is where the query is executed, hence it's the bottleneck of the queries

    http://jsonapi.org/format/#fetching-pagination

    A server MAY choose to limit the number of resources returned
    in a response to a subset (“page”) of the whole set available.
    A server MAY provide links to traverse a paginated data set (“pagination links”).
    Pagination links MUST appear in the links object that corresponds
    to a collection. To paginate the primary data, supply pagination links
    in the top-level links object. To paginate an included collection
    returned in a compound document, supply pagination links in the
    corresponding links object.

    The following keys MUST be used for pagination links:

    first: the first page of data
    last: the last page of data
    prev: the previous page of data
    next: the next page of data

    We use page[offset] and page[limit], where
    offset is the number of records to offset by prior to returning resources

    :param object_query: SQLAalchemy query object
    :param SAFRSObject: optional
    :return: links, instances, count
    """

    def get_link(count, limit):
        result = SAFRSObject._s_url if SAFRSObject else ""
        ignore_args = "page[offset]", "page[limit]"
        result += "?" + "&".join(
            [f"{k}={v}" for k, v in request.args.items() if k not in ignore_args] + [f"page[offset]={count}&page[limit]={limit}"]
        )
        return result

    try:
        page_offset = int(get_request_param("page_offset"))
        limit = int(get_request_param("page_limit", get_config("DEFAULT_PAGE_LIMIT")))
    except ValueError:
        raise ValidationError("Pagination Value Error")

    if limit <= 0:
        limit = 1
    if limit > get_config("MAX_PAGE_LIMIT"):
        limit = get_config("MAX_PAGE_LIMIT")
    if page_offset <= 0:
        page_offset = 0
    if page_offset > get_config("MAX_PAGE_OFFSET"):
        page_offset = get_config("MAX_PAGE_OFFSET")
    page_base = int(page_offset / limit) * limit

    # Counting may take > 1s for a table with millions of records, depending on the storage engine :|
    # Make it configurable
    # With mysql innodb we can use following to retrieve the count:
    # select TABLE_ROWS from information_schema.TABLES where TABLE_NAME = 'TableName';
    if isinstance(object_query, (list, sqlalchemy.orm.collections.InstrumentedList)):
        count = len(object_query)
    elif SAFRSObject is None:  # for backwards compatibility, ie. when not passed as an arg to paginate()
        count = object_query.count()
    else:
        count = SAFRSObject._s_count()
    if count is None:
        try:
            count = object_query.count()
        except Exception as exc:
            # May happen for custom types, for ex. the psycopg2 extension
            safrs.log.warning(f"Can't get count for {SAFRSObject} ({exc})")
            count = -1

        if count > get_config("MAX_TABLE_COUNT"):
            safrs.log.warning("Large table count detected, performance may be impacted, consider using '_s_count'")

    first_args = (0, limit)
    last_args = (int(int(count / limit) * limit), limit)  # round down
    self_args = (page_base if page_base <= last_args[0] else last_args[0], limit)
    next_args = (page_offset + limit, limit) if page_offset + limit <= last_args[0] else last_args
    prev_args = (page_offset - limit, limit) if page_offset > limit else first_args

    links = {
        "first": get_link(*first_args),
        "self": get_link(page_offset, limit),  # cfr. request.url
        "last": get_link(*last_args),
        "prev": get_link(*prev_args),
        "next": get_link(*next_args),
    }

    if last_args == self_args:
        del links["last"]
    if first_args == self_args:
        del links["first"]
    if next_args == last_args:
        del links["next"]
    if prev_args == first_args:
        del links["prev"]

    if isinstance(object_query, (list, sqlalchemy.orm.collections.InstrumentedList)):
        instances = object_query[page_offset : page_offset + limit]
    elif isinstance(object_query, dict):
        # (might happen when using a custom filter)
        instances = object_query
    elif not hasattr(object_query, "offset"):
        # in case the query is emulated
        instances = []
    else:
        try:
            res_query = object_query.offset(page_offset).limit(limit)
            instances = res_query.all()
        except OverflowError:
            raise ValidationError("Pagination Overflow Error")
        except sqlalchemy.exc.CompileError as exc:  # pragma: no cover
            # dirty workaround for when no sort parameter is provided in case of mssql offset/limit queries
            # error msg:
            #     "MSSQL requires an order_by when using an OFFSET or a non-simple LIMIT clause"
            safrs.log.warning(f"{exc} / Add a valid sort= URL parameter")
            if "MSSQL requires an order_by" in str(exc) and SAFRSObject and SAFRSObject.id_type.primary_keys:
                pk = SAFRSObject.id_type.primary_keys[0]
                res_query = object_query.order_by(getattr(SAFRSObject, pk)).offset(page_offset).limit(limit)
                instances = res_query.all()
            else:
                raise GenericError(f"{exc}")
        except Exception as exc:
            raise GenericError(f"{exc}")

    return links, instances, count


def jsonapi_format_response(data=None, meta=None, links=None, errors=None, count=None, include=None):
    """
    Create a response dict according to the json:api schema spec
    :param data : the objects that will be serialized
    :return: jsonapi formatted dictionary
    """
    limit = get_request_param("page_limit", get_config("MAX_PAGE_LIMIT"))
    try:
        limit = int(limit)
    except ValueError:
        raise ValidationError("page[limit] error")
    if meta is None:
        meta = {}

    meta["limit"] = limit
    meta["count"] = meta["total"] = count

    jsonapi = dict(version="1.0")
    result = dict(data=data)

    if errors:
        result["errors"] = errors
    if meta:
        result["meta"] = meta
    if jsonapi:
        result["jsonapi"] = jsonapi
    if links:
        result["links"] = links

    result["included"] = safrs.base.Included

    return result



# Some custom types for db columns and jsonapi id coding
import uuid
import datetime
import hashlib
import re
import json
import safrs
from sqlalchemy.types import PickleType, String
from sqlalchemy.types import TypeDecorator, BLOB
from .errors import ValidationError
from .util import classproperty


STRIP_SPECIAL = r"[^\w|%|:|/|-|_\-_\. ]"


class SAFRSID:
    """
    This class creates a jsonapi "id" from the classes PKs
    In case of a composite PK, the pks are joined with the delimiter
    eg.
    pkA = 1, pkB = 2, delimiter = '_' => jsonapi_id = '1_2'

    If you want to create a custom id_type, you can subclass SAFRSID
    """

    primary_keys = ["id"]
    columns = None
    delimiter = "_"

    def __new__(cls, id=None):

        if id is None:
            return cls.gen_id()
        else:
            return cls.validate_id(id)

    @classmethod
    def gen_id(cls):
        """
        Generate a jsonapi id
        """
        # This is the case if an autoincrement id is expected:
        if cls.columns and len(cls.columns) == 1 and cls.columns[0].type.python_type == int:
            return None

        # Some dialects support UUID
        # Maybe we should use db.UUID() instead
        return str(uuid.uuid4())

    @classmethod
    def validate_id(cls, id):
        """
        Validate a given id (eg. check if it's a valid uuid, email etc.)
        """
        result = id
        if len(cls.columns) == 1:
            try:
                result = cls.columns[0].type.python_type(id)
            except Exception:
                raise ValidationError(f"Invalid id: '{id}'.")
        else:
            pass
            # safrs.log.debug("ID Validation not implemented for {}".format(cls))

        return result

    @property
    def name(self):
        """
        name
        """
        return self.delimiter.join(self.primary_keys)

    @classmethod
    def get_id(cls, obj):
        """
        Retrieve the id string derived from the pks of obj
        """
        if cls.columns and len(cls.columns) > 1:
            values = [str(getattr(obj, pk.name)) for pk in cls.columns]
            return cls.delimiter.join(values)

        return getattr(obj, cls.primary_keys[0])

    @classmethod
    def get_pks(cls, jsonapi_id):
        """
        Convert the jsonapi_id string to a pk dict
        in case the PK is composite it consists of PKs joined by cls.delimiter
        Note: there may be an issue when the cls.delimiter is contained in an id
        :return: primary key dict
        """
        if len(cls.columns) == 1:
            values = [jsonapi_id]
        else:
            values = str(jsonapi_id).split(cls.delimiter)
        if len(values) != len(cls.columns):
            columns = [c.name for c in cls.columns]
            raise ValidationError(f"PK values ({values}) do not match columns ({columns})")
        result = dict()
        for pk_col, val in zip(cls.columns, values):
            if not val:
                if pk_col.type.python_type == int:
                    val = 0
            try:
                col_name = str(pk_col.name)
                result[col_name] = pk_col.type.python_type(val)
            except (ValueError, TypeError):  # pragma: no cover
                # This may happen when val is empty '' or
                # if when the swagger doc is generated with default uuids
                # todo: fix
                if pk_col.default:
                    result[col_name] = pk_col.default
                else:
                    result[col_name] = ""
            except Exception as exc:  # pragma: no cover
                safrs.log.warning(f"PK Error: {exc}")
                result[col_name] = ""

        return result

    @classproperty
    def column_names(cls):
        """
        :return: a list of columns names of this id type
        """
        return [c.name for c in cls.columns]

    @classmethod
    def extract_pks(cls, kw_dict):
        """
        Extract the primary keys from kw_dict (these are the kwargs passed to SAFRSBase.new())
        In case of composite keys we construct the jsonapi_id by using the delimiter to join the values
        :return: primary keys dict
        """
        pks = {k: str(kw_dict[k]) for k in cls.column_names}
        id = cls.delimiter.join(pks.values())
        return cls.get_pks(id)

    @classmethod
    def sample_id(cls, obj):
        if cls.columns and len(cls.columns) == 1 and cls.columns[0].type.python_type == int:
            return 0
        sample = None
        try:
            sample = obj.query.first()
        except Exception as exc:
            safrs.log.debug(exc)
            pass
        if sample:
            return sample.jsonapi_id

        return "jsonapi_id_string"


def get_id_type(cls, Super=SAFRSID, delimiter="_"):
    """
    get_id_type
    """
    columns = [col for col in cls.__table__.columns if col.primary_key]
    primary_keys = [col.name for col in columns]
    delimiter = getattr(cls, "delimiter", "_")
    id_type_class = type(cls.__name__ + "_ID", (Super,), {"primary_keys": primary_keys, "columns": columns, "delimiter": delimiter})
    return id_type_class


class SAFRSSHA256HashID(SAFRSID):  # pragma: no cover
    """
    SAFRSSHA256HashID class for a hash based id
    """

    @classmethod
    def gen_id(cls):
        """
        Create a hash based on the current time
        This is just an example
        Not cryptographically secure and might cause collisions!
        """
        now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f").encode("utf-8")
        return hashlib.sha256(now).hexdigest()

    @classmethod
    def validate_id(cls, _id):
        """
        validate_id
        """
        # todo
        return _id


class JSONType(PickleType):  # pragma: no cover
    """
    JSON DB type is used to store JSON objects in the database
    """

    impl = BLOB

    def __init__(self, *args, **kwargs):

        # kwargs['pickler'] = json
        super(JSONType, self).__init__(*args, **kwargs)

    def process_bind_param(self, value, dialect):

        if value is not None:
            value = json.dumps(value, ensure_ascii=True)
        return value

    def process_result_value(self, value, dialect):

        if value is not None:
            value = json.loads(value)
        return value


class SafeString(TypeDecorator):  # pragma: no cover
    """
    DB String Type class strips special chars when bound
    """

    impl = String(767)

    def __init__(self, *args, **kwargs):

        super(SafeString, self).__init__(*args, **kwargs)

    def process_bind_param(self, value, dialect):

        if value is not None:
            result = re.sub(STRIP_SPECIAL, "_", value)
            if str(result) != str(value):
                # log.warning('({}) Replaced {} by {}'.format(self, value, result))
                pass
        else:
            result = value

        return result


class UUIDType(TypeDecorator):  # pragma: no cover
    """
    UUIDType
    """

    impl = String(40)

    def __init__(self, *args, **kwargs):

        super(UUIDType, self).__init__(*args, **kwargs)

    @staticmethod
    def process_bind_param(value, dialect):

        try:
            uuid.UUID(value, version=4)
        except Exception as exc:
            raise ValidationError(f"UUID Validation Error {value} ({exc})")

        return value



import datetime
import safrs
import sqlalchemy


def parse_attr(column, attr_val):
    """
    Parse the supplied `attr_val` so it can be saved in the SQLAlchemy `column`

    :param attr: SQLAlchemy column
    :param attr_val: jsonapi attribute value
    :return: processed value
    """
    if attr_val is None and column.default:
        attr_val = column.default.arg
        return attr_val

    if attr_val is None:
        return attr_val

    if getattr(column, "python_type", None):
        # It's possible for a column to specify a custom python_type to use for deserialization
        attr_val = column.python_type(attr_val)

    try:
        column.type.python_type
    except NotImplementedError as exc:
        """
        This happens when a custom type has been implemented, in which case the user/dev should know how to handle it:
        override this method and implement the parsing
        https://docs.python.org/2/library/exceptions.html#exceptions.NotImplementedError :
        In user defined base classes, abstract methods should raise this exception when they require derived classes to override the method.
        => simply return the attr_val for user-defined classes
        """
        safrs.log.debug(exc)
        return attr_val

    # skip type coercion on JSON columns, since they could be anything
    if type(column.type) is sqlalchemy.sql.sqltypes.JSON:
        return attr_val

    """
        Parse datetime and date values for some common representations
        If another format is used, the user should create a custom column type or custom serialization
    """
    if attr_val and column.type.python_type == datetime.datetime:
        date_str = str(attr_val)
        try:
            if "." in date_str:
                # str(datetime.datetime.now()) => "%Y-%m-%d %H:%M:%S.%f"
                attr_val = datetime.datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S.%f")
            else:
                # JS datepicker format
                attr_val = datetime.datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S")
        except (NotImplementedError, ValueError) as exc:
            safrs.log.warning(f'Invalid datetime.datetime {exc} for value "{attr_val}"')
            attr_val = datetime.datetime.now()
    elif attr_val and column.type.python_type == datetime.date:
        try:
            attr_val = datetime.datetime.strptime(str(attr_val), "%Y-%m-%d")
        except (NotImplementedError, ValueError) as exc:
            safrs.log.warning(f'Invalid datetime.date {exc} for value "{attr_val}"')
            attr_val = datetime.datetime.now()
    elif attr_val and column.type.python_type == datetime.time:  # pragma: no cover (todo)
        try:
            date_str = str(attr_val)
            if "." in date_str:
                # str(datetime.datetime.now()) => "%H:%M:%S.%f"
                attr_val = datetime.datetime.strptime(str(attr_val), "%H:%M:%S.%f").time()
            else:
                # JS datepicker format
                attr_val = datetime.datetime.strptime(str(attr_val), "%H:%M:%S").time()
        except (NotImplementedError, ValueError, TypeError) as exc:
            safrs.log.warning(f'Invalid datetime.time {exc} for value "{attr_val}"')
            attr_val = column.type.python_type()
    else:
        attr_val = column.type.python_type(attr_val)

    return attr_val



# flask_restful_swagger2 API subclass
from http import HTTPStatus
import logging
import werkzeug
from flask_restful import abort, Resource
from flask_restful.representations.json import output_json
from flask_restful.utils import OrderedDict
from flask_restful.utils import cors
from flask_restful_swagger_2 import Api as FRSApiBase
from flask_restful_swagger_2 import validate_definitions_object, parse_method_doc
from flask_restful_swagger_2 import validate_path_item_object, Schema
from flask_restful_swagger_2 import extract_swagger_path, Extractor, ValidationError as FRSValidationError
from flask import request
from functools import wraps
import safrs
from .swagger_doc import swagger_doc, swagger_method_doc, default_paging_parameters
from .swagger_doc import parse_object_doc, swagger_relationship_doc, get_http_methods
from .errors import JsonapiError, SystemValidationError, GenericError
from .config import get_config
from .json_encoder import SAFRSJSONEncoder
from ._safrs_relationship import SAFRSRelationshipObject
from sqlalchemy.orm.interfaces import MANYTOONE
from flask import current_app, Response
import json
import yaml

HTTP_METHODS = ["GET", "POST", "PATCH", "DELETE", "PUT"]
DEFAULT_REPRESENTATIONS = [("application/vnd.api+json", output_json)]


class SAFRSAPI(FRSApiBase):
    """
    Subclass of the flask_restful_swagger API class where we add the expose_object method
    this method creates an API endpoint for the SAFRSBase object and corresponding swagger
    documentation
    """

    _operation_ids = {}
    _custom_swagger = {}
    _als_resources = []

    def __init__(
        self,
        app,
        host="localhost",
        port=5000,
        prefix="",
        description="SAFRSAPI",
        json_encoder=SAFRSJSONEncoder,
        swaggerui_blueprint=True,
        **kwargs,
    ):
        """
        http://jsonapi.org/format/#content-negotiation-servers
        Servers MUST send all JSON:API data in response documents with
        the header Content-Type: application/vnd.api+json without any media type parameters.

        Servers MUST respond with a 415 Unsupported Media Type status code if
        a request specifies the header Content-Type: application/vnd.api+json with any media type parameters.

        Servers MUST respond with a 406 Not Acceptable status code if
        a request’s Accept header contains the JSON:API media type and
        all instances of that media type are modified with media type parameters.
        """

        self._custom_swagger = kwargs.pop("custom_swagger", {})
        kwargs["default_mediatype"] = "application/vnd.api+json"
        safrs.SAFRS(app, prefix=prefix, json_encoder=json_encoder, swaggerui_blueprint=swaggerui_blueprint, **kwargs)
        # the host shown in the swagger ui
        # this host may be different from the hostname of the server and
        # sometimes we don't want to show the port (eg when proxied)
        # in that case the port may be None
        if port:
            host = f"{host}:{port}"

        super().__init__(
            app,
            api_spec_url=kwargs.pop("api_spec_url", "/swagger"),
            host=host,
            description=description,
            prefix=prefix,
            base_path=prefix,
            **kwargs,
        )
        self.init_app(app)
        self.representations = OrderedDict(DEFAULT_REPRESENTATIONS)
        self.update_spec()

    def update_spec(self):
        """
        :param custom_swagger: swagger spec to be added to the swagger.json
        """
        _swagger_doc = self.get_swagger_doc()
        safrs.dict_merge(_swagger_doc, self._custom_swagger)

    def expose_object(self, safrs_object, url_prefix="", **properties):
        """This methods creates the API url endpoints for the SAFRObjects
        :param safrs_object: SAFSBase subclass that we would like to expose
        :param url_prefix: url prefix
        :param properties: additional flask-restful properties

        creates a class of the form

        @api_decorator
        class Class_API(SAFRSRestAPI):
            SAFRSObject = safrs_object

        add the class as an api resource to /SAFRSObject and /SAFRSObject/{id}

        tablename/collectionname: safrs_object._s_collection_name, e.g. "Users"
        classname: safrs_object.__name__, e.g. "User"
        """
        if not current_app:
            safrs.log.error("Working outside of app context!")
        rest_api = safrs_object._rest_api  # => SAFRSRestAPI

        properties["SAFRSObject"] = safrs_object
        properties["http_methods"] = safrs_object.http_methods
        safrs_object.url_prefix = url_prefix
        endpoint = safrs_object.get_endpoint()

        # tags indicate where in the swagger hierarchy the endpoint will be shown
        tags = [safrs_object._s_collection_name]

        # Expose the methods first
        self.expose_methods(url_prefix, tags, safrs_object, properties)

        # Expose the collection: Create the class and decorate it
        api_class_name = f"{safrs_object._s_type}_API"  # name for dynamically generated classes
        RESOURCE_URL_FMT = get_config("RESOURCE_URL_FMT")  # configurable resource collection url formatter
        url = RESOURCE_URL_FMT.format(url_prefix, safrs_object._s_collection_name)
        swagger_decorator = swagger_doc(safrs_object)
        api_class = api_decorator(type(api_class_name, (rest_api,), properties), swagger_decorator)

        safrs.log.info(f"Exposing {safrs_object._s_collection_name} on {url}, endpoint: {endpoint}")
        self.add_resource(api_class, url, endpoint=endpoint, methods=["GET", "POST"])

        INSTANCE_URL_FMT = get_config("INSTANCE_URL_FMT")
        url = INSTANCE_URL_FMT.format(url_prefix, safrs_object._s_collection_name, safrs_object.__name__)
        endpoint = safrs_object.get_endpoint(type="instance")

        # Expose the instances
        safrs.log.info(f"Exposing {safrs_object._s_type} instances on {url}, endpoint: {endpoint}")
        api_class = api_decorator(type(api_class_name + "_i", (rest_api,), properties), swagger_decorator)
        self.add_resource(api_class, url, endpoint=endpoint, methods=["GET", "PATCH", "DELETE"])

        object_doc = parse_object_doc(safrs_object)
        object_doc["name"] = safrs_object._s_collection_name
        self._swagger_object["tags"].append(object_doc)

        for relationship in safrs_object._s_relationships.values():
            self.expose_relationship(relationship, url, tags, properties)

        # add newly created schema references to the "definitions"
        for def_name, definition in Schema._references.items():
            if self._swagger_object["definitions"].get(def_name):
                continue
            try:
                validate_definitions_object(definition.properties)
            except Exception as exc:  # pragma: no cover
                safrs.log.warning(f"Failed to validate {definition}:{exc}")
                continue
            self._swagger_object["definitions"][def_name] = {"properties": definition.properties}

        self.update_spec()
        self._als_resources.append(safrs_object)

    def expose_methods(self, url_prefix, tags, safrs_object, properties):
        """
        Expose the safrs "documented_api_method" decorated methods
        :param url_prefix: api url prefix
        :param tags: swagger tags
        :return: None
        """
        rpc_api = safrs_object._rpc_api  # => SAFRSJSONRPCAPI
        api_methods = safrs_object._s_get_jsonapi_rpc_methods()
        for api_method in api_methods:
            method_name = api_method.__name__
            api_method_class_name = f"method_{safrs_object._s_class_name}_{method_name}"
            if (
                isinstance(safrs_object.__dict__.get(method_name, None), (classmethod, staticmethod))
                or getattr(api_method, "__self__", None) is safrs_object
            ):
                # method is a classmethod or static method, make it available at the class level
                CLASSMETHOD_URL_FMT = get_config("CLASSMETHOD_URL_FMT")
                url = CLASSMETHOD_URL_FMT.format(url_prefix, safrs_object._s_collection_name, method_name)
            else:
                # expose the method at the instance level
                INSTANCEMETHOD_URL_FMT = get_config("INSTANCEMETHOD_URL_FMT")
                url = INSTANCEMETHOD_URL_FMT.format(url_prefix, safrs_object._s_collection_name, safrs_object._s_object_id, method_name)

            ENDPOINT_FMT = get_config("ENDPOINT_FMT")
            endpoint = ENDPOINT_FMT.format(url_prefix, safrs_object._s_collection_name + "." + method_name)
            swagger_decorator = swagger_method_doc(safrs_object, method_name, tags)
            properties.update({"method_name": method_name, "http_methods": safrs_object.http_methods})
            api_class = api_decorator(type(api_method_class_name, (rpc_api,), properties), swagger_decorator)
            meth_name = safrs_object._s_class_name + "." + api_method.__name__
            safrs.log.info(f"Exposing method {meth_name} on {url}, endpoint: {endpoint}")
            self.add_resource(api_class, url, endpoint=endpoint, methods=get_http_methods(api_method), jsonapi_rpc=True)

    def expose_relationship(self, relationship, url_prefix, tags, properties):
        """
        Expose a relationship tp the REST API:
        A relationship consists of a parent and a target class
        creates a class of the form

        @api_decorator
        class Parent_X_target_API(SAFRSRestRelationshipAPI):
            SAFRSObject = safrs_object

        add the class as an api resource to /SAFRSObject and /SAFRSObject/{id}

        :param relationship: relationship
        :param url_prefix: api url prefix
        :param tags: swagger tags
        :return: None
        """
        # safrs_object is the target class, if this is not a SAFRSBase class, then we shouldn't expose it
        # the _s_expose attribute indicates we're dealing with a SAFRSBase instance
        # if the relationship is not an sql sqlalchemy.orm.relationships.RelationshipProperty instance
        # then we should have defined the _target
        target_object = relationship.mapper.class_
        relationship_api = target_object._relationship_api  # => SAFRSRestRelationshipAPI
        if not getattr(target_object, "_s_expose", False):  # todo: add test
            safrs.log.debug(f"Not exposing {target_object}")
            return

        API_CLASSNAME_FMT = "{}_X_{}_API"  # api class name for generated relationship classes
        rel_name = relationship.key
        parent_class = relationship.parent.class_
        parent_name = parent_class.__name__

        # Name of the endpoint class
        RELATIONSHIP_URL_FMT = get_config("RELATIONSHIP_URL_FMT")
        api_class_name = API_CLASSNAME_FMT.format(parent_name, rel_name)
        url = RELATIONSHIP_URL_FMT.format(url_prefix, rel_name)

        ENDPOINT_FMT = get_config("ENDPOINT_FMT")
        endpoint = ENDPOINT_FMT.format(url_prefix, rel_name)

        # Relationship object
        decorators = set(
            getattr(parent_class, "custom_decorators", [])
            + getattr(parent_class, "decorators", [])
            + getattr(relationship, "decorators", [])
        )
        rel_object = type(
            f"{parent_name}.{rel_name}",  # Name of the class we're creating here
            (SAFRSRelationshipObject,),
            {
                "relationship": relationship,
                # Merge the relationship decorators from the classes
                # This makes things really complicated!!!
                # TODO: simplify this by creating a proper superclass
                "custom_decorators": list(decorators),
                "parent": parent_class,
                "_target": target_object,
            },
        )

        properties["SAFRSObject"] = rel_object
        properties["http_methods"] = target_object.http_methods
        swagger_decorator = swagger_relationship_doc(rel_object, tags)
        api_class = api_decorator(type(api_class_name, (relationship_api,), properties), swagger_decorator)

        # Expose the relationship for the parent class:
        # GET requests to this endpoint retrieve all item ids
        safrs.log.info(f"Exposing relationship {rel_name} on {url}, endpoint: {endpoint}")
        # Check if there are custom http methods specified
        methods = getattr(relationship, "http_methods", parent_class.http_methods)
        self.add_resource(api_class, url, endpoint=endpoint, methods=methods, relationship=relationship)

        try:
            target_object_id = target_object._s_object_id
        except Exception as exc:
            safrs.log.exception(exc)
            safrs.log.error(f"No object id for {target_object}")
            target_object_id = target_object.__name__

        if target_object == parent_class:
            # Avoid having duplicate argument ids in the url:
            # append a 2 in case of a self-referencing relationship
            # todo : test again
            target_object_id += "2"

        # Expose the relationship for <string:targetId>, this lets us
        # query and delete the class relationship properties for a given
        # target id
        # nb: this is not really documented in the jsonapi spec, remove??
        url = (RELATIONSHIP_URL_FMT + "/<string:{}>").format(url_prefix, rel_name, target_object_id)
        endpoint = f"{url_prefix}api.{rel_name}Id"

        safrs.log.info(f"Exposing {parent_name} relationship {rel_name} on {url}, endpoint: {endpoint}")
        self.add_resource(
            api_class, url, relationship=rel_object.relationship, endpoint=endpoint, methods=["GET", "PATCH", "DELETE"], deprecated=True
        )

    @staticmethod
    def get_resource_methods(resource, ordered_methods=None):
        """
        :param ordered_methods:
        :return: the http methods from the SwaggerEndpoint and SAFRS Resources,
        in the order specified by ordered_methods
        """
        if ordered_methods is None:
            ordered_methods = HTTP_METHODS
        om = ordered_methods
        safrs_object = getattr(resource, "SAFRSObject", None)
        if safrs_object:
            om = [m.upper() for m in safrs_object.http_methods if m.upper() in ordered_methods]

        resource_methods = [m.lower() for m in ordered_methods if m in resource.methods and m.upper() in om]
        return resource_methods

    def add_resource(self, resource, *urls, **kwargs):
        """
        This method is partly copied from flask_restful_swagger_2/__init__.py

        Changed because we don't need path id examples when there's no {id} in the path.
        We also have to filter out the unwanted parameters
        """
        relationship = kwargs.pop("relationship", False)  # relationship object
        SAFRS_INSTANCE_SUFFIX = get_config("OBJECT_ID_SUFFIX") + "}"

        path_item = {}
        self._add_oas_resource_definitions(resource, path_item)
        is_jsonapi_rpc = kwargs.pop("jsonapi_rpc", False)  # check if the exposed method is a jsonapi_rpc method
        deprecated = kwargs.pop("deprecated", False)  # deprecated functionality: still working but not shown in swagger

        # this loop builds the swagger for the specified url(s) by adding it to
        # self._swagger_object["paths"][swagger_url], so if the loop continues,
        # there will be no swagger
        # usually there will only be one url, but flask_restful add_resource does support multiple urls
        for url in urls:
            if deprecated:
                # functionality still works, but there will be no swagger
                continue
            if not url.startswith("/"):  # pragma: no cover
                raise SystemValidationError("paths must start with a /")

            swagger_url = extract_swagger_path(url)
            # exposing_instance tells us whether we're exposing an instance (as opposed to a collection)
            exposing_instance = swagger_url.strip("/").endswith(SAFRS_INSTANCE_SUFFIX)
            if relationship:
                exposing_instance = relationship.direction == MANYTOONE
            for method in self.get_resource_methods(resource):
                if kwargs.get("methods", None) and method.upper() not in [m.upper() for m in kwargs.get("methods", [])]:
                    # only use the
                    path_item.pop(method, None)
                    continue

                if method == "post" and exposing_instance:
                    # POSTing to an instance isn't jsonapi-compliant (https://jsonapi.org/format/#crud-creating-client-ids)
                    # "A server MUST return 403 Forbidden in response to an
                    # unsupported request to create a resource with a client-generated ID"
                    # the method has already been added before, remove it & continue
                    path_item.pop(method, None)
                    continue

                method_doc = path_item.get(method)
                if not method_doc:
                    continue

                # exposed objects and methods may specify `collection_summary` in the method docstring yaml
                collection_summary = method_doc.pop("collection_summary", method_doc.get("summary", None))
                if not exposing_instance and collection_summary:
                    method_doc["summary"] = collection_summary

                method_doc["operationId"] = self._get_operation_id(path_item.get(method).get("summary", ""))

                self._add_oas_req_params(resource, path_item, method, exposing_instance, is_jsonapi_rpc, swagger_url)
                self._add_oas_references(resource.SAFRSObject, path_item, method, exposing_instance, relationship)

                try:  # pragma: no cover
                    validate_path_item_object(path_item)
                except FRSValidationError as exc:
                    safrs.log.exception(exc)
                    safrs.log.critical(f"Validation failed for {path_item}")
                    exit(1)

            self._swagger_object["paths"][swagger_url] = path_item
            # Check whether we manage to convert to json
            try:
                json.dumps(self._swagger_object)
            except Exception:  # pragma: no cover
                safrs.log.critical("Json encoding failed")

        # disable API methods that were not set by the SAFRSObject
        for http_method in HTTP_METHODS:
            hm = http_method.lower()
            if hm not in self.get_resource_methods(resource):
                setattr(resource, hm, lambda x: ({}, HTTPStatus.METHOD_NOT_ALLOWED))
        # pylint: disable=bad-super-call
        super(FRSApiBase, self).add_resource(resource, *urls, **kwargs)

    def _add_oas_req_params(self, resource, path_item, method, exposing_instance, is_jsonapi_rpc, swagger_url):
        """
        Add the request parameters to the swagger (filter, sort)
        """
        method_doc = path_item[method]
        parameters = []
        for parameter in method_doc.get("parameters", []):
            object_id = "{%s}" % parameter.get("name")
            if method == "get":
                # Get the jsonapi included resources, ie the exposed relationships
                param = resource.get_swagger_include()
                parameters.append(param)

                # Get the jsonapi fields[], ie the exposed attributes/columns
                # only required for collections though
                param = resource.get_swagger_fields()
                parameters.append(param)

            #
            # Add the sort, filter parameters to the swagger doc when retrieving a collection
            #
            if method == "get" and not (exposing_instance or is_jsonapi_rpc):
                # limit parameter specifies the number of items to return
                parameters += default_paging_parameters()
                param = resource.get_swagger_sort()
                parameters.append(param)
                parameters += list(resource.get_swagger_filters())

            if not (parameter.get("in") == "path" and object_id not in swagger_url) and parameter not in parameters:
                # Only if a path param is in path url then we add the param
                parameters.append(parameter)

        unique_params = OrderedDict()  # rm duplicates
        for param in parameters:
            unique_params[param["name"]] = param
        method_doc["parameters"] = list(unique_params.values())
        path_item[method] = method_doc

    def _add_oas_references(self, safrs_object, path_item, method, exposing_instance, relationship):
        """
        substitute the swagger references in the response objects
        references are created and added to the safrs_object.swagger_models in swagger_doc

        the params are
        :param safrs_object:
        :param path_item:

        """
        inst_ref = None
        coll_ref = None

        if getattr(safrs_object, "swagger_models", {}).get("instance"):
            # instance reference
            inst_ref = safrs_object.swagger_models["instance"].reference()

        if getattr(safrs_object, "swagger_models", {}).get("collection"):
            # collection reference
            coll_ref = safrs_object.swagger_models["collection"].reference()

        if not inst_ref and not coll_ref:
            return

        method_doc = path_item[method]
        response = method_doc.get("responses", {})
        if "200" in response:
            # add the "example" response schema references
            if exposing_instance:
                response["200"]["schema"] = inst_ref
            elif coll_ref:
                response["200"]["schema"] = coll_ref
            response["200"].pop("type", None)

        if "201" in response:
            if method == "post":
                # Posting to a collection returns the instance (except when using bulk post, but this isn't shown atm)
                response["201"]["schema"] = inst_ref
            elif exposing_instance:
                # patching an instance
                response["201"]["schema"] = inst_ref
            elif coll_ref:
                # patching a
                response["201"]["schema"] = coll_ref
            response["201"]["schema"].pop("type", None)

    def _add_oas_resource_definitions(self, resource, path_item):
        """
        add the resource method schema references to the swagger "definitions"
        :param resource:
        :param path_item:
        """
        definitions = {}

        for method in self.get_resource_methods(resource):
            if not method.upper() in HTTP_METHODS:
                continue
            f = getattr(resource, method, None)
            if not f:
                continue

            operation = getattr(f, "__swagger_operation_object", None)
            if operation:
                operation, definitions_ = Extractor.extract(operation)
                path_item[method] = operation
                definitions.update(definitions_)
                summary = parse_method_doc(f, operation)
                if summary:
                    operation["summary"] = summary.split("<br/>")[0]

        try:
            validate_definitions_object(definitions)
        except FRSValidationError:
            safrs.log.critical(f"Validation failed for {definitions}")
            exit()

        self._swagger_object["definitions"].update(definitions)

    @classmethod
    def _get_operation_id(cls, summary):
        """
        :param summary:
        """
        summary = "".join(c for c in summary if c.isalnum())
        if summary not in cls._operation_ids:
            cls._operation_ids[summary] = 0
        else:
            cls._operation_ids[summary] += 1
        return f"{summary}_{cls._operation_ids[summary]}"

    def expose_als_schema(self, api_root="/api", schema_loc="/als_schema"):
        """
        Generate the resource specification for apilogicserver
        """
        resources = {}
        result = {"resources": resources, "api_root": api_root}
        for resource in self._als_resources:
            resource_data = {"type": resource._s_type, "label": None}
            attributes = []
            for name, attr_def in resource._s_jsonapi_attrs.items():
                attr = {}
                attr["name"] = name
                # column["type"] = col.python_type
                attributes.append(attr)
            resource_data["attributes"] = attributes
            resource_data["perPage"] = 10
            relations = []
            for rel_name, rel in resource._s_relationships.items():
                relation = {}
                relation["name"] = rel_name
                relation["resource"] = str(rel.target.key)
                relation["fks"] = [str(c.key) for c in rel._calculated_foreign_keys]
                relation["direction"] = "toone" if rel.direction == MANYTOONE else "tomany"
                relations.append(relation)
            resource_data["tab_groups"] = relations

            resources[resource._s_collection_name] = resource_data

        class ApiSchema(Resource):
            def get(self):
                if request.args.get("yaml"):
                    return Response(yaml.dump(result), content_type="text/yaml")
                return result

        self.add_resource(ApiSchema, "/als_schema")
        return json.dumps(result, indent=4)


def api_decorator(cls, swagger_decorator):
    """Decorator for the API views:
        - add swagger documentation ( swagger_decorator )
        - add cors
        - add generic exception handling

    We couldn't use inheritance because the rest method decorator
    references the cls.SAFRSObject which isn't known

    :param cls: The class that will be decorated (e.g. SAFRSRestAPI, SAFRSRestRelationshipAPI)
    :param swagger_decorator: function that will generate the swagger
    :return: decorated class
    """

    cors_domain = get_config("cors_domain")
    cls.http_methods = {}  # holds overridden http methods, note: cls also has "methods" set, but it's not related to this
    for method_name in [
        "patch",
        "post",
        "delete",
        "get",
        "put",
        "options",
    ]:  # HTTP methods, "put isn't used by us but may be used by a hacky developer"
        method = getattr(cls, method_name, None)
        if not method:
            continue

        decorated_method = method
        # if the SAFRSObject has a custom http method decorator, use it
        # e.g. SAFRSObject.get
        custom_method = getattr(cls.SAFRSObject, method_name, None)
        if custom_method and callable(custom_method):
            decorated_method = custom_method
            # keep the default method as parent_<method_name>, e.g. parent_get
            parent_method = getattr(cls, method_name)
            cls.http_methods[method_name] = parent_method

        # Add cors
        if cors_domain is not None:
            decorated_method = cors.crossdomain(origin=cors_domain)(decorated_method)
        # Add exception handling
        decorated_method = http_method_decorator(decorated_method)
        setattr(decorated_method, "SAFRSObject", cls.SAFRSObject)

        if method_name != "options":
            try:
                # Add swagger documentation
                decorated_method = swagger_decorator(decorated_method)
            except RecursionError:  # pragma: no cover
                # Got this error when exposing WP DB, TODO: investigate where it comes from
                safrs.log.error(f"Failed to generate documentation for {cls} {decorated_method} (Recursion Error)")

            except Exception as exc:
                safrs.log.exception(exc)
                safrs.log.error(f"Failed to generate documentation for {decorated_method}")

            # The user can add custom decorators
            # Apply the custom decorators, specified as class variable list
            for custom_decorator in set(getattr(cls.SAFRSObject, "custom_decorators", []) + getattr(cls.SAFRSObject, "decorators", [])):
                # update_wrapper(custom_decorator, decorated_method)
                swagger_operation_object = getattr(decorated_method, "__swagger_operation_object", {})
                decorated_method = custom_decorator(decorated_method)
                decorated_method.__swagger_operation_object = swagger_operation_object

        setattr(cls, method_name, decorated_method)
    return cls


def http_method_decorator(fun):
    """Decorator for the supported jsonapi HTTP methods (get, post, patch, delete)
    - commit the database
    - convert all exceptions to a JSON serializable GenericError

    This method will be called for all requests
    :param fun:
    :return: wrapped fun
    """

    @wraps(fun)
    def method_wrapper(*args, **kwargs):
        """Wrap the method and perform error handling
        :param *args:
        :param **kwargs:
        :return: result of the wrapped method
        """
        safrs_exception = None
        try:
            if not request.is_jsonapi and fun.__name__ not in ["get", "head", "options", "delete"]:  # pragma: no cover
                # reuire jsonapi content type for requests to these routes
                raise GenericError(HTTPStatus.UNSUPPORTED_MEDIA_TYPE.description, HTTPStatus.UNSUPPORTED_MEDIA_TYPE.value)
            result = fun(*args, **kwargs)
            safrs.DB.session.commit()
            return result

        except JsonapiError as exc:
            safrs.log.exception(exc)
            safrs_exception = exc

        except werkzeug.exceptions.NotFound as exc:
            status_code = 404
            safrs_exception = exc

        except Exception as exc:
            safrs.log.exception(exc)
            safrs_exception = exc
            if safrs.log.getEffectiveLevel() > logging.DEBUG:
                safrs_exception.message = "Logging Disabled"
            else:
                safrs_exception.message = str(exc)

        status_code = getattr(safrs_exception, "status_code", 500)
        api_code = getattr(safrs_exception, "api_code", status_code)
        title = getattr(safrs_exception, "message", "")
        detail = getattr(safrs_exception, "detail", title)

        safrs.DB.session.rollback()
        safrs.log.error(detail)
        errors = dict(title=title, detail=detail, code=api_code)
        abort(status_code, errors=[errors])

    return method_wrapper



"""
python3 setup.py sdist
twine upload dist/*
"""

from distutils.core import setup

def safrs_setup():
    with open("requirements.txt", "rt") as fp:
        install_requires = fp.read().strip().split("\n")

    version = "2.12.2"

    setup(
        name="safrs",
        packages=["safrs"],
        version=version,
        license="MIT",
        description="safrs : SqlAlchemy Flask-Restful Swagger2",
        long_description=open("README.rst").read(),
        author="Thomas Pollet",
        author_email="thomas.pollet@gmail.com",
        url="https://github.com/thomaxxl/safrs",
        download_url="https://github.com/thomaxxl/safrs/archive/{}.tar.gz".format(version),
        keywords=["SqlAlchemy", "Flask", "REST", "Swagger", "JsonAPI", "OpenAPI"],
        python_requires=">=3.6, !=3.0.*, !=3.1.*, !=3.2.*, <4",
        install_requires=install_requires,
        classifiers=[
            "Development Status :: 3 - Alpha",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Intended Audience :: Developers",
            "Framework :: Flask",
            "Topic :: Software Development :: Libraries",
            "Environment :: Web Environment",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.6",
        ],
        extras_require = { 'admin' : ['Flask-Admin>=1.5.8', 'Flask-Cors>=3.0.9'] , 'db2api' : ['inflect==5.0.2', 'Flask-Cors>=3.0.9'] }
    )

safrs_setup() # pragma: no cover



#  This file contains jsonapi-related flask-restful "Resource" objects:
#  - SAFRSRestAPI for exposed database instances and collections
#  - SAFRSRestRelationshipAPI for exposed database relationships
#  - SAFRSRestMethodAPI for exposed jsonapi_rpc methods
#
# Configuration parameters:
# - endpoint
#
# to do:
# - expose canonical endpoints
# - move all swagger formatting to swagger_doc
#
# Some linting errors to ignore
# pylint: disable=redefined-builtin,invalid-name, line-too-long, protected-access, no-member, too-many-lines
# pylint: disable=fixme, logging-format-interpolation
#
import safrs
import sqlalchemy
import sqlalchemy.orm.dynamic
import sqlalchemy.orm.collections
from flask import jsonify, make_response as flask_make_response, url_for, request
from flask_restful_swagger_2 import Resource as FRSResource
from http import HTTPStatus
from sqlalchemy.orm.interfaces import MANYTOONE
from .swagger_doc import is_public
from .errors import ValidationError, NotFoundError
from .jsonapi_formatting import jsonapi_filter_query, jsonapi_filter_list, jsonapi_sort, jsonapi_format_response, paginate
from .jsonapi_filters import get_swagger_filters


def make_response(*args, **kwargs):
    """
    Customized flask-restful make_response
    """
    response = flask_make_response(*args, **kwargs)
    if request.is_jsonapi:
        # Only use "application/vnd.api+json" if the client sent this with the request
        response.headers["Content-Type"] = "application/vnd.api+json"
    return response


class Resource(FRSResource):
    """
    Superclass for the exposed endpoints
    * Collections and instances : SAFRSRestAPI
    * Relationships : SAFRSRestRelationshipAPI
    * RPC methods : SAFRSJSONRPCAPI
    """

    # SAFRSObject: the class that will be returned when a http method is invoked
    # Flask views will need to set this to the SQLAlchemy safrs.DB.Model class
    SAFRSObject = None
    # relationship target in SAFRSRestRelationshipAPI, identical to self.SAFRSObject in SAFRSRestAPI
    target = None
    # Swagger filter spec
    get_swagger_filters = get_swagger_filters

    def head(self, *args, **kwargs):
        """
        HTTP HEAD
        """
        _super = super()
        if hasattr(_super, "head"):
            response = _super.head(*args, **kwargs)
        else:
            response = make_response()
        return response

    def options(self, *args, **kwargs):
        """
        HTTP OPTIONS
        """
        _super = super()
        if hasattr(_super, "options"):
            response = _super.options(*args, **kwargs)
        else:
            response = make_response()
        return response

    def _parse_target_data(self, target_data):
        """
        Validate the jsonapi payload for patch requests (to self.target):
        - the payload must contain "id" and "type" keys.
        - the type must match the target type
        - an object with the specified id must exist

        :param target_data: jsonapi instance payload
        :return: sqla/safrs orm instance
        """
        if not isinstance(target_data, dict):
            raise ValidationError(f"Invalid data type {target_data}")
        target_id = target_data.get("id", None)
        if target_id is None:
            raise ValidationError(f"no target id {target_data}")
        target_type = target_data.get("type")
        if not target_id:
            raise ValidationError("Invalid id in data", HTTPStatus.FORBIDDEN)
        if not target_type:
            raise ValidationError("Invalid type in data", HTTPStatus.FORBIDDEN)
        if target_type != self.target._s_type:
            raise ValidationError(f"Invalid type {target_type} != {self.target._s_type}", HTTPStatus.FORBIDDEN)
        target = self.target.get_instance(target_id)
        if not target:
            raise ValidationError(f"invalid target id {target_id}")
        return target

    @classmethod
    def get_swagger_include(cls):
        """
        :return: JSON:API "include" query string swagger spec
        """
        default_include = ",".join(cls.SAFRSObject._s_relationships.keys())

        param = {
            "default": default_include,
            "type": "string",
            "name": "include",
            "in": "query",
            "format": "string",
            "required": False,
            "description": f"{cls.SAFRSObject._s_class_name} relationships to include (csv)",
        }
        return param

    @classmethod
    def get_swagger_fields(cls):
        """
        :return: JSON:API fields[] swagger spec (the model instance fields to be included)
        """
        attr_list = list(cls.SAFRSObject._s_jsonapi_attrs.keys())
        # Add the fields query string swagger
        # todo: get the columns of the target
        param = {
            "default": ",".join(attr_list),
            "type": "string",
            "name": f"fields[{cls.SAFRSObject._s_class_name}]",
            "in": "query",
            "format": "string",
            "required": False,
            "description": f"{cls.SAFRSObject._s_class_name} fields to include (csv)",
        }
        return param

    @classmethod
    def get_swagger_sort(cls):
        """
        :return: JSON:API sort swagger spec (the collection sort key)
        """
        attr_list = list(cls.SAFRSObject._s_jsonapi_attrs.keys()) + ["id"]

        param = {
            "default": ",".join(attr_list),
            "type": "string",
            "name": "sort",
            "in": "query",
            "format": "string",
            "required": False,
            "description": "Sort order",
        }
        return param


class SAFRSRestAPI(Resource):
    """
    Flask webservice wrapper for the underlying Resource Object:
    an sqla db model (SAFRSBase subclass : cls.SAFRSObject)

    This class implements HTTP Methods (get, post, put, delete, ...) and helpers

    http://jsonapi.org/format/#document-resource-objects
    A resource object MUST contain at least the following top-level members:
    - id
    - type

    In addition, a resource object MAY contain any of these top-level members:

    attributes: an attributes object representing some of the resource’s data.
    relationships: a relationships object describing relationships
                    between the resource and other JSON API resources.
    links: a links object containing links related to the resource.
    meta: a meta object containing non-standard meta-information
            about a resource that can not be represented as an attribute or relationship.

    e.g.
    {
        "id": "1f1c0e90-9e93-4242-9b8c-56ac24e505e4",
        "type": "car",
        "attributes": {
            "color": "red"
        },
        "relationships": {
            "driver": {
                "data": {
                    "id": "55550e90-9e93-4242-9b8c-56ac24e505e5",
                    "type": "person"
                }
            }
    }

    A resource object’s attributes and its relationships are collectively called its “fields”.
    """

    default_order = None  # used by sqla order_by
    object_id = None

    def __init__(self, *args, **kwargs):
        """
        - object_id is the function used to create the url parameter name
        (eg "User" -> "UserId" )
        - this parameter is used in the swagger endpoint spec,
        eg. /Users/{UserId} where the UserId parameter is the id of
        the underlying SAFRSObject.
        """
        self._s_object_id = self.SAFRSObject._s_object_id
        self.target = self.SAFRSObject

    def get(self, **kwargs):
        """
        summary : Retrieve {class_name} instance
        description : Retrieve {class_name} from {collection_name}
        responses :
            200 :
                description : Request fulfilled, document follows
            403 :
                description : Forbidden
            404 :
                description : Not Found
        ---
        HTTP GET: return instances
        If no id is given: return all instances
        If an id is given, get an instance by id

        id is specified by self._s_object_id, f.i. {UserId}

        http://jsonapi.org/format/#document-top-level

        A JSON object MUST be at the root of every JSON API request
        and response containing data. This object defines a document’s “top level”.

        A document MUST contain at least one of the following top-level members:
        - data: the document’s “primary data”
        - errors: an array of error objects
        - meta: a meta object that contains non-standard meta-information.

        A document MAY contain any of these top-level members:
        - jsonapi: an object describing the server’s implementation
        - links: a links object related to the primary data.
        - included: an array of resource objects that are related
        to the primary data and/or each other (“included resources”).
        """
        data = None
        meta = {}
        errors = None
        links = None

        if self._s_object_id in kwargs:
            # Retrieve a single instance
            id = kwargs[self._s_object_id]
            instance = self.SAFRSObject.get_instance(id)
            data = instance
            links = {"self": instance._s_url}
            if request.url != instance._s_url:
                links["related"] = request.url
            count = 1
            meta.update(dict(instance_meta=instance._s_meta()))
        else:
            # retrieve a collection, filter and sort
            instances = self.SAFRSObject._s_get()
            instances = jsonapi_sort(instances, self.SAFRSObject)
            links, data, count = paginate(instances, self.SAFRSObject)

        # format the response: add the included objects
        result = jsonapi_format_response(data, meta, links, errors, count)
        return make_response(jsonify(result))

    def patch(self, **kwargs):
        """
        summary : Update {class_name}
        description: Update {class_name} attributes
        responses:
            200 :
                description : Request fulfilled, document follows
            202 :
                description : Accepted
            204 :
                description : No Content
            403:
                description : Forbidden
            404 :
                description : Not Found
            409 :
                description : Conflict
        ---
        https://jsonapi.org/format/#crud-updating-responses
        Update the object with the specified id
        """
        id = kwargs.get(self._s_object_id, None)

        payload = request.get_jsonapi_payload()
        if not isinstance(payload, dict):
            raise ValidationError("Invalid Object Type")

        data = payload.get("data")
        if id is None and isinstance(data, list):
            # Bulk patch request
            for item in data:
                if not isinstance(item, dict):
                    raise ValidationError("Invalid Data Object")
                instance = self._patch_instance(item)
            response = make_response(jsonify({}), HTTPStatus.ACCEPTED)

        elif not data or not isinstance(data, dict):
            raise ValidationError("Invalid Data Object")
        elif id is None:
            raise ValidationError("Invalid ID")
        else:
            instance = self._patch_instance(data, id)
            # object id is the endpoint parameter, for example "UserId" for a User SAFRSObject
            obj_args = {instance._s_object_id: instance.jsonapi_id}
            # Retrieve the jsonapi encoded object and return it to the client
            obj_data = self.get(**obj_args)
            response = make_response(obj_data, HTTPStatus.OK)
            # Set the Location header to the newly created object
            response.headers["Location"] = url_for(self.endpoint, **obj_args)

        return response

    def _patch_instance(self, data, id=None):
        """
        Update the inst
        :param data: jsonapi payload
        :param id: jsonapi id
        :return: instance
        """
        # validate the jsonapi id in the url path and convert it to a database id
        path_id = self.SAFRSObject.id_type.validate_id(id)
        # Check that the id in the body is equal to the id in the url
        body_id = data.get("id", None)
        if body_id is None:
            raise ValidationError("No ID in body")

        body_id = self.SAFRSObject.id_type.validate_id(body_id)
        if path_id is not None and path_id != body_id:
            raise ValidationError(f"Invalid ID {type(path_id)} {path_id} != {type(body_id)} {body_id}")

        attributes = data.get("attributes", {})
        attributes["id"] = body_id
        # Create the object instance with the specified id and json data
        # If the instance (id) already exists, it will be updated with the data
        instance = self._parse_target_data(data)
        if not instance:
            raise ValidationError("No instance with ID")
        instance._s_patch(**attributes)

        return instance

    def post(self, **kwargs):
        """
        summary : Create {class_name}
        responses :
            403:
                description : Forbidden
            201:
                description: Created
            202:
                description : Accepted
            404:
                description : Not Found
            409:
                description : Conflict
        ---
        http://jsonapi.org/format/#crud-creating
        Creating Resources
        A resource can be created by sending a POST request to a URL
        that represents a collection of resources.
        The request MUST include a single resource object as primary data.
        The resource object MUST contain at least a type member.

        If a relationship is provided in the relationships member of the resource object,
        its value MUST be a relationship object with a data member.
        The value of this key represents the linkage the new resource is to have.

        Response:
        403: This implementation does not accept client-generated IDs
        201: Created
        202: Accepted
        404: Not Found
        409: Conflict

        Location Header identifying the location of the newly created resource
        Body : created object

        TODO:
        409 Conflict
          A server MUST return 409 Conflict when processing a POST request
          to create a resource with a client-generated ID that already exists.
          A server MUST return 409 Conflict when processing a POST request
          in which the resource object’s type is not among the type(s) that
          constitute the collection represented by the endpoint.
          A server SHOULD include error details and provide enough
          information to recognize the source of the conflict.
        """
        payload = request.get_jsonapi_payload()
        id = kwargs.get(self._s_object_id, None)
        if id is not None:
            # POSTing to an instance isn't jsonapi-compliant (https://jsonapi.org/format/#crud-creating-client-ids)
            # to do: modify Allow header
            raise ValidationError(f"POSTing to instance is not allowed {self}", status_code=HTTPStatus.METHOD_NOT_ALLOWED)

        # Create a new instance of the SAFRSObject
        data = payload.get("data")
        if data is None:
            raise ValidationError("Request contains no data")
        if isinstance(data, list):
            # http://springbot.github.io/json-api/extensions/bulk/
            # We should verify that the bulk extension is requested
            # Accept it by default now
            if not request.is_bulk:
                safrs.log.warning("Client sent a bulk POST but did not specify the bulk extension")
            instances = []
            for item in data:
                instance = self._create_instance(item)
                instances.append(instance)
            resp_data = jsonify({"data": instances})
            location = None
        else:
            instance = self._create_instance(data)
            # object_id is the endpoint parameter, for example "UserId" for a User SAFRSObject
            obj_args = {instance._s_object_id: instance.jsonapi_id}
            # Retrieve the object json and return it to the client
            resp_data = self.get(**obj_args)
            location = f"{url_for(self.endpoint)}{instance.jsonapi_id}"

        response = make_response(resp_data, HTTPStatus.CREATED)
        # Set the Location header to the newly created object(s)
        if location:
            response.headers["Location"] = location

        return response

    def _create_instance(self, data):
        """
        Create an instance with the
        :param data: dictionary with {"type": ... , "attributes": ...}
        :return: created instance
        """
        if not isinstance(data, dict):
            raise ValidationError("Data is not a dict object")

        obj_type = data.get("type", None)
        if not obj_type or not obj_type == self.SAFRSObject._s_type:
            raise ValidationError(f"Invalid type member: {obj_type} != {self.SAFRSObject._s_type}")

        attributes = data.get("attributes", {})
        if self.SAFRSObject.allow_client_generated_ids:
            client_generated_id = data.get("id", None)
            attributes["id"] = client_generated_id
        elif "id" in data:
            safrs.log.warning(f"Client-generated ids are not allowed for {self.SAFRSObject}")

        relationships = data.get("relationships", {})

        instance = self.SAFRSObject._s_post(**attributes, **relationships)

        return instance

    def delete(self, **kwargs):
        """
        summary: Delete {class_name} from {collection_name}
        responses :
            202 :
                description: Accepted
            204 :
                description: Request fulfilled, nothing follows
            200 :
                description: Success
            403 :
                description: Forbidden
            404 :
                description: Not Found

        ---
        Delete an object by id or by filter

        http://jsonapi.org/format/1.1/#crud-deleting:
        Responses :
            202 : Accepted
            If a deletion request has been accepted for processing,
            but the processing has not been completed by the time the server
            responds, the server MUST return a 202 Accepted status code.

            204 No Content
            A server MUST return a 204 No Content status code if a deletion
            request is successful and no content is returned.

            200 OK
            A server MUST return a 200 OK status code if a deletion request
            is successful and the server responds with only top-level meta data.

            404 NOT FOUND
            A server SHOULD return a 404 Not Found status code
            if a deletion request fails due to the resource not existing.
        """
        id = kwargs.get(self._s_object_id, None)

        if not id:
            # This endpoint shouldn't be exposed so this code is not reachable
            raise ValidationError("", status_code=HTTPStatus.METHOD_NOT_ALLOWED)

        instance = self.SAFRSObject.get_instance(id)
        instance._s_delete()

        return make_response(jsonify({}), HTTPStatus.NO_CONTENT)


class SAFRSRestRelationshipAPI(Resource):
    """
    Flask webservice wrapper for the underlying sqla relationships db model

    The endpoint url is of the form
    "/Parents/{ParentId}/children/{ChildId}"
    (cfr RELATIONSHIP_URL_FMT in API.expose_relationship)
    where "children" is the relationship attribute of the parent

    3 types of relationships (directions) exist in the sqla orm:
    MANYTOONE ONETOMANY MANYTOMANY

    Following attributes are set on this class:
        - SAFRSObject: the sqla object which has been set with the type
         constructor in expose_relationship
        - source_class: class of the parent ( e.g. Parent, __tablename__ : Parents )
        - target : class of the child
        - rel_name : name of the relationship ( e.g. children )
        - parent_object_id : url parameter name of the parent ( e.g. {ParentId} )
        - child_object_id : url parameter name of the child ( e.g. {ChildId} )

    http://jsonapi.org/format/#crud-updating-relationships

    Updating To-Many Relationships
    A server MUST respond to PATCH, POST, and DELETE requests to a URL
    from a to-many relationship link as described below.

    For all request types, the body MUST contain a data member
    whose value is an empty array or an array of resource identifier objects.

    If a client makes a PATCH request to a URL from a to-many relationship link,
    the server MUST either completely replace every member of the relationship,
    return an appropriate error response if some resources can not be
    found or accessed, or return a 403 Forbidden response if complete
    replacement is not allowed by the server.
    """

    SAFRSObject = None

    # pylint: disable=unused-argument
    def __init__(self, *args, **kwargs):
        """
        Initialize the relationship references:
        - relationship : sqla relationship
        -
        """
        self.relationship = self.SAFRSObject.relationship
        self.source_class = self.SAFRSObject.relationship.parent.class_
        self.target = self.SAFRSObject.relationship.mapper.class_
        self.rel_name = self.SAFRSObject.relationship.key
        # The object_ids are the ids in the swagger path e.g {FileId}
        self.parent_object_id = self.source_class._s_object_id
        self.child_object_id = self.target._s_object_id

        if self.parent_object_id == self.child_object_id:
            # see expose_relationship: if a relationship consists of
            # two same objects, the object_id should be different (i.e. append "2")
            self.child_object_id += "2"

    # Retrieve relationship data
    def get(self, **kwargs):
        """
        summary : Retrieve {child_name} from {cls.relationship.key}
        description : Retrieve {child_name} items from the {parent_name} {cls.relationship.key} "{direction}" relationship
        ---
        https://jsonapi.org/format/#fetching-relationships

        Retrieve a relationship or list of relationship member ids

        http://jsonapi.org/format/#fetching-relationships-responses :
        A server MUST respond to a successful request to fetch a
        relationship with a 200 OK response.The primary data in the response
        document MUST match the appropriate value for resource linkage.
        The top-level links object MAY contain self and related links,
        as described above for relationship objects.
        """
        _, relation = self.parse_args(**kwargs)
        child_id = kwargs.get(self.child_object_id)
        errors = {}
        count = 1
        meta = {}
        data = None

        if relation is None:
            # child may have been deleted
            return "Not Found", HTTPStatus.NOT_FOUND
        elif self.SAFRSObject.relationship.direction == MANYTOONE:
            data = instance = relation
            links = {"self": request.url}
            if request.url != instance._s_url:
                links["related"] = instance._s_url
            meta.update(dict(instance_meta=instance._s_meta()))
        elif child_id:
            data = child = self.target.get_instance(child_id)
            links = {"self": child._s_url}
            # If {ChildId} is passed in the url, return the child object
            # there's a difference between to-one and -to-many relationships:
            if isinstance(relation, safrs.SAFRSBase) and child != relation:
                raise NotFoundError()
            elif child not in relation:
                raise NotFoundError()
            else:
                links = {"self": request.url, "related": child._s_url}
        elif isinstance(relation, sqlalchemy.orm.collections.InstrumentedList):
            instances = jsonapi_filter_list(relation)
            instances = jsonapi_sort(instances, self.target)
            links, data, count = paginate(instances, self.target)
            count = len(data)
        else:
            # lazy='dynamic' relationships
            instances = jsonapi_filter_query(relation, self.target)
            instances = jsonapi_sort(instances, self.target)
            links, data, count = paginate(instances, self.target)

        result = jsonapi_format_response(data, meta, links, errors, count)
        return make_response(jsonify(result))

    # Relationship patching
    def patch(self, **kwargs):
        """
        summary : Update {cls.relationship.key}
        description : Update the {parent_name} {cls.relationship.key} "{direction}" relationship
        responses:
            200 :
                description : Accepted
            204 :
                description : No Content
            403:
                description : Forbidden
            404 :
                description : Not Found
            409 :
                description : Conflict
        ----
        Update or create a relationship child item
        to be used to create or update one-to-many mappings
        but also works for many-to-many etc.

        # Updating To-One Relationships

        http://jsonapi.org/format/#crud-updating-to-one-relationships:
        A server MUST respond to PATCH requests to a URL
        from a to-one relationship link as described below

        The PATCH request MUST include a top-level member named data containing one of:
        a resource identifier object corresponding to the new related resource.
        null, to remove the relationship.
        """
        changed = False
        parent, relation = self.parse_args(**kwargs)
        payload = request.get_jsonapi_payload()
        data = payload.get("data")
        relation = getattr(parent, self.rel_name)
        obj_args = {self.parent_object_id: parent.jsonapi_id}

        if isinstance(data, dict):
            # https://jsonapi.org/format/#crud-updating-to-one-relationships
            # server MUST respond to PATCH requests to a URL from a to-one relationship link as described below.
            #   The PATCH request MUST include a top-level member named data containing one of:
            #   a resource identifier object corresponding to the new related resource.
            #   null, to remove the relationship.

            if self.SAFRSObject.relationship.direction != MANYTOONE:
                raise ValidationError("Provide a list to PATCH a TOMANY relationship")
            child = self._parse_target_data(data)
            if getattr(parent, self.rel_name) != child:
                # change the relationship, i.e. add the child
                setattr(parent, self.rel_name, child)
                obj_args[self.child_object_id] = child.jsonapi_id
                changed = True

        elif isinstance(data, list) and not self.SAFRSObject.relationship.direction == MANYTOONE:
            """
            http://jsonapi.org/format/#crud-updating-to-many-relationships

            If a client makes a PATCH request to a URL from a to-many relationship link,
            the server MUST either completely replace every member of the relationship,
            return an appropriate error response if some resourcescan not be found
            or accessed, or return a 403 Forbidden response if complete
            replacement is not allowed by the server.
            """
            # first remove all items, then append the new items
            # if the relationship has been configured with lazy="dynamic"
            # then it is a subclass of AppenderBaseQuery and
            # we should empty the relationship by setting it to []
            # otherwise it is an instance of InstrumentedList and we have to empty it
            # (we could loop all items but this is slower for large collections)
            tmp_rel = []
            for child_data in data:
                child = self._parse_target_data(child_data)
                tmp_rel.append(child)

            if isinstance(relation, sqlalchemy.orm.collections.InstrumentedList):
                relation[:] = tmp_rel
            else:
                setattr(parent, self.rel_name, tmp_rel)

        elif data is None and self.SAFRSObject.relationship.direction == MANYTOONE:
            # { data : null } //=> clear the relationship
            child = getattr(parent, self.SAFRSObject.relationship.key)
            if child:
                pass
            setattr(parent, self.rel_name, None)
        else:
            raise ValidationError(
                f'Invalid data object type "{type(data)}" for this "{self.SAFRSObject.relationship.direction}"" relationship'
            )

        # Create the patch response
        # https://jsonapi.org/format/#crud-updating-responses
        # 200 OK
        # If a server accepts an update but also changes the resource(s) in ways other than those specified by the request
        # (for example, updating the updated-at attribute or a computed sha), it MUST return a 200 OK response. The response
        # document MUST include a representation of the updated resource(s) as if a GET request was made to the request URL.
        # A server MUST return a 200 OK status code if an update is successful, the client’s current attributes remain up to date,
        # and the server responds only with top-level meta data. In this case the server MUST NOT include a representation of the updated resource(s).
        # 204 No Content
        # If an update is successful and the server doesn’t update any attributes besides those provided, the server MUST return
        # either a 200 OK status code and response document (as described above) or a 204 No Content status code with no response document.

        if data is None:
            # item removed from relationship => 202 accepted
            data, status_code = {}, HTTPStatus.NO_CONTENT
        elif changed:
            return self.get(**obj_args)
        else:
            # Nothing changed, reflect the data
            data, status_code = {"data": data}, HTTPStatus.OK

        return make_response(jsonify(data), status_code)

    # Adding items to a relationship
    def post(self, **kwargs):
        """
        summary: Add {child_name} items to {cls.relationship.key}
        description : Add {child_name} items to the {parent_name} {cls.relationship.key} "{direction}" relationship
        responses :
            202:
                description : Accepted
            204:
                description : No Content
            404:
                description : Not Found
            409:
                description : Conflict
        ---
        Add a child to a relationship
        202 Accepted
        If a relationship update request has been accepted for processing, but the processing has not
        been completed by the time the server responds, the server MUST return a 202 Accepted status code.

        204 No Content
        A server MUST return a 204 No Content status code if an update is successful and the representation
        of the resource in the request matches the result.
        """
        parent, relation = self.parse_args(**kwargs)
        payload = request.get_jsonapi_payload()
        data = payload.get("data", None)

        if data is None:
            raise ValidationError("Invalid POST payload (no data)")

        if self.SAFRSObject.relationship.direction == MANYTOONE:
            # https://jsonapi.org/format/#crud-updating-to-one-relationships
            # We should only use patch to update a relationship
            child_data = data
            if isinstance(child_data, list):
                raise ValidationError(
                    """
                    Invalid data payload: MANYTOONE relationship can only hold a single item,
                    please provide a dictionary object
                    """,
                    HTTPStatus.FORBIDDEN,
                )
            if child_data:
                child = self._parse_target_data(child_data)
                setattr(parent, self.rel_name, child)
            data = {"data": child}
            status_code = HTTPStatus.OK
        else:  # direction is TOMANY => append the items to the relationship
            for child_data in data:
                child = self._parse_target_data(child_data)
                if child not in relation:
                    relation.append(child)
            data = {}
            status_code = HTTPStatus.NO_CONTENT

        # we can return result too but it's not necessary per the spec
        return make_response(jsonify(data), status_code)

    def delete(self, **kwargs):
        """
        summary : Delete {child_name} from {cls.relationship.key}
        description : Delete {child_name} items from the {parent_name} {cls.relationship.key} "{direction}" relationship
        responses:
            202 :
                description: Accepted
            204 :
                description: Request fulfilled, nothing follows
            200 :
                description: Success
            403 :
                description: Forbidden
            404 :
                description: Not Found
        ----
        Remove an item from a relationship
        """
        # pylint: disable=unused-variable
        # (parent is unused)
        parent, relation = self.parse_args(**kwargs)

        # No child id=> delete specified items from the relationship
        payload = request.get_jsonapi_payload()
        if not isinstance(payload, dict):
            raise ValidationError("Invalid Object Type")
        data = payload.get("data")

        if self.SAFRSObject.relationship.direction == MANYTOONE:
            # https://jsonapi.org/format/#crud-updating-to-one-relationships
            # We should only use patch to update
            # previous versions incorrectly implemented the jsonapi spec for updating manytoone relationships
            # keep things backwards compatible for now
            child = data
            if isinstance(data, list):
                if data and isinstance(data[0], dict):
                    # invalid, try to fix it by deleting the firs item from the list
                    safrs.log.warning("Invalid Payload to delete from MANYTOONE relationship")
                    data = data[0]
            if not isinstance(data, dict):
                raise ValidationError("Invalid data payload")
            child_id = data.get("id", None)
            child_type = data.get("type", None)

            if child_id is None or child_type is None:
                raise ValidationError("Invalid data payload", HTTPStatus.FORBIDDEN)

            if child_type != self.target._s_type:
                raise ValidationError("Invalid type", HTTPStatus.FORBIDDEN)

            child = self.target.get_instance(child_id)
            if child == relation and getattr(parent, self.rel_name, None) == child:
                # Delete the item from the many-to-one relationship
                delattr(parent, self.rel_name)
            else:
                safrs.log.warning("child not in relation")

        else:
            # https://jsonapi.org/format/#crud-updating-to-many-relationships
            children = data
            if not isinstance(data, list) or not children:
                raise ValidationError("Invalid data payload")
            for child in children:
                child_id = child.get("id", None)
                child_type = child.get("type", None)

                if not child_id or not child_type:
                    raise ValidationError("Invalid data payload", HTTPStatus.FORBIDDEN)

                if child_type != self.target._s_type:
                    raise ValidationError("Invalid type", HTTPStatus.FORBIDDEN)

                child = self.target.get_instance(child_id)
                if child in relation:
                    relation.remove(child)
                else:
                    safrs.log.warning(f"Item with id {child_id} not in relation")

        return make_response(jsonify({}), HTTPStatus.NO_CONTENT)

    def parse_args(self, **kwargs):
        """
        Parse relationship args
        An error is raised if the parent doesn't exist.

        :return: parent, child, relation
        """
        parent_id = kwargs.get(self.parent_object_id, None)
        if parent_id is None:
            raise ValidationError("Invalid Parent Id")

        parent = self.source_class.get_instance(parent_id)
        relation = getattr(parent, self.rel_name)

        return parent, relation


class SAFRSJSONRPCAPI(Resource):
    """
    Route wrapper for the underlying SAFRSBase jsonapi_rpc

    Only HTTP POST is supported
    """

    SAFRSObject = None  # Flask views will need to set this to the SQLAlchemy safrs.DB.Model class
    method_name = None

    def __init__(self, *args, **kwargs):
        """
        -object_id is the function used to create the url parameter name
        (eg "User" -> "UserId" )
        -this parameter is used in the swagger endpoint spec,
        eg. /Users/{UserId} where the UserId parameter is the id of the underlying SAFRSObject.
        """
        self._s_object_id = self.SAFRSObject._s_object_id
        self.target = self.SAFRSObject

    def post(self, **kwargs):
        """
        summary : call
        responses :
            403:
                description :
            201:
                description: Created
            202:
                description : Accepted
            403 :
                description : Forbidden
            404:
                description : Not Found
            409:
                description : Conflict
        ---
        HTTP POST: apply actions, return 200 regardless.
        The actual jsonapi_rpc method may return other codes
        """
        id = kwargs.get(self._s_object_id, None)

        if id is not None:
            instance = self.SAFRSObject.get_instance(id)
            if not instance:
                # If no instance was found this means the user supplied
                # an invalid ID
                raise ValidationError("Invalid ID")
        else:
            # No ID was supplied, apply method to the class itself
            instance = self.SAFRSObject

        method = getattr(instance, self.method_name, None)

        if not method:
            # Only call methods for Campaign and not for superclasses (e.g. safrs.DB.Model)
            raise ValidationError(f'Invalid method "{self.method_name}"')
        if not is_public(method):
            raise ValidationError("Method is not public")

        args = dict(request.args)
        if getattr(method, "valid_jsonapi", False):
            payload = request.get_jsonapi_payload()
            if payload:
                args = payload.get("meta", {}).get("args", {})
        else:
            args = request.get_json()

        return self._create_rpc_response(method, args)

    def get(self, **kwargs):
        """
        responses :
            404 :
                description : Not Found
            403 :
                description : Forbidden

        ---
        """
        id = kwargs.get(self._s_object_id, None)

        if id is not None:
            instance = self.SAFRSObject.get_instance(id)
            if not instance:
                # If no instance was found this means the user supplied
                # an invalid ID
                raise ValidationError("Invalid ID")

        else:
            # No ID was supplied, apply method to the class itself
            instance = self.SAFRSObject

        method = getattr(instance, self.method_name, None)

        if not method:
            # Only call methods for Campaign and not for superclasses (e.g. safrs.DB.Model)
            raise ValidationError(f'Invalid method "{self.method_name}"')
        if not is_public(method):
            raise ValidationError("Method is not public")

        args = dict(request.args)
        return self._create_rpc_response(method, args)

    def _create_rpc_response(self, method, args):

        safrs.log.debug(f"method {self.method_name} args {args}")
        result = method(**args)

        if isinstance(result, safrs.SAFRSFormattedResponse):
            response = result
        elif getattr(method, "valid_jsonapi", None) is False:
            response = result
        else:
            response = {"meta": {"result": result}}

        return make_response(jsonify(response), HTTPStatus.OK)



"""
    jsonapi_attr: custom jsonapi attributes
"""
from sqlalchemy.ext.hybrid import hybrid_property
from .swagger_doc import parse_object_doc

JSONAPI_ATTR_TAG = "_s_is_jsonapi_attr"


class jsonapi_attr(hybrid_property):
    """
    hybrid_property type: sqlalchemy.orm.attributes.create_proxied_attribute.<locals>.Proxy
    """

    def __init__(self, *args, **kwargs):
        """
        :param attr: `SAFRSBase` attribute that should be exposed by the jsonapi
        :return: jsonapi attribute decorator

        set `swagger_type` and `default` to customize the swagger
        """
        setattr(self, JSONAPI_ATTR_TAG, True)  # checked by is_jsonapi_attr()

        if args:
            # called when the app starts
            attr = args[0]
            obj_doc = parse_object_doc(attr)
            if isinstance(obj_doc, dict):
                for k, v in obj_doc.items():
                    setattr(self, k, v)
        else:
            # the "default" kwarg may have been added by the obj_doc but we no longer
            # need it (and it causes an exception)
            kwargs.pop("default", None)
        super().__init__(*args, **kwargs)

    def getter(self, fget):
        """
        Provide a decorator that defines a getter method.
        """

        return self._copy(fget=fget)

    def setter(self, fset):
        """
        Provide a decorator that defines a setter method.
        """

        return self._copy(fset=fset)


def is_jsonapi_attr(attr):
    """
    :param attr: `SAFRSBase` `jsonapi_attr` decorated attribute
    :return: boolean
    """
    return getattr(attr, JSONAPI_ATTR_TAG, False) is True



# safrs to json encoding

import datetime
import decimal
import json
from flask.json import JSONEncoder
from sqlalchemy.ext.declarative import DeclarativeMeta
from uuid import UUID
import safrs
from .config import is_debug
from .base import SAFRSBase, Included
from .jsonapi_formatting import jsonapi_format_response


class SAFRSFormattedResponse:
    """
    Custom response object
    """

    # pylint: disable=too-few-public-methods
    data = None
    meta = None
    errors = None
    result = None
    response = None

    def __init__(self, *args, **kwargs):
        """
        :param data:
        :param meta:
        :param links:
        :param errors:
        :param count:
        """
        self.response = jsonapi_format_response(*args, **kwargs)

    def to_dict(self):  # pragma: no cover
        """
        create the response payload that will be sent to the browser
        :return: dict or None
        """
        if self.response is not None:
            return self.response

        if self.meta is not None:
            return self.meta

        if self.result is not None:
            return {"meta": {"result": self.result}}

        return None


class SAFRSJSONEncoder(JSONEncoder):
    """
    Encodes safrs objs (SAFRSBase subclasses)
    """

    # pylint: disable=too-many-return-statements,logging-format-interpolation
    # pylint: disable=arguments-differ,protected-access,method-hidden
    def default(self, obj, **kwargs):
        """
        override the default json encoding
        :param obj: object to be encoded
        :return: encoded/serizlaized object
        """
        if obj is Included:
            return Included.encode()
        if isinstance(obj, Included):
            result = obj.encode()
            return result
        if isinstance(obj, SAFRSBase):
            result = obj._s_jsonapi_encode()
            return result
        if isinstance(obj, datetime.datetime):
            return obj.isoformat(" ")
        if isinstance(obj, (datetime.date, datetime.time)):
            return obj.isoformat()
        if isinstance(obj, set):
            return list(obj)
        if isinstance(obj, SAFRSFormattedResponse):
            return obj.to_dict()
        if isinstance(obj, UUID):  # pragma: no cover
            return str(obj)
        if isinstance(obj, decimal.Decimal):  # pragma: no cover
            return float(obj)
        if isinstance(obj, bytes):  # pragma: no cover
            if obj == b"":
                return ""
            safrs.log.debug("SAFRSJSONEncoder: serializing bytes obj")
            return obj.hex()

        # We shouldn't get here in a normal setup
        # getting here means we already abused safrs... and we're no longer jsonapi compliant
        if not is_debug():  # pragma: no cover
            # only continue if in debug mode
            safrs.log.warning(f'JSON Encoding Error: Unknown object type "{type(obj)}" for {obj}')
            return {"error": "SAFRSJSONEncoder invalid object"}

        if isinstance(obj, DeclarativeMeta):  # pragma: no cover
            return self.sqla_encode(obj)

        return self.ghetto_encode(obj)

    @staticmethod
    def ghetto_encode(obj):  # pragma: no cover
        """
        if everything else failed, try to encode the public obj attributes
        i.e. those attributes without a _ prefix
        :param obj: object to be encoded
        :return: encoded/serizlaized object
        """
        try:
            result = {}
            for k, v in vars(obj).items():
                if not k.startswith("_"):
                    if isinstance(v, (int, float)) or v is None:
                        result[k] = v
                    else:
                        result[k] = str(v)
        except TypeError:
            result = str(obj)
        return result

    # pragma: no cover
    @staticmethod
    def sqla_encode(obj):  # pragma: no cover
        """
        encode an SQLAlchemy object
        :param obj: sqlalchemy object to be encoded
        :return: encoded object
        """
        fields = {}
        for field in [x for x in dir(obj) if not x.startswith("_") and x != "metadata"]:
            data = obj.__getattribute__(field)
            try:
                # this will raise an exception if the data can't be serialized
                fields[field] = json.dumps(data)
            except TypeError:
                fields[field] = str(data)

        # a json-encodable dict
        return fields
